from datetime import timedelta
from typing import Literal

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="fastlife_")

    fastlife_route_prefix: str = Field(default="/_fl")
    template_search_path: str = Field(default="fastlife:templates")
    registry_class: str = Field(default="fastlife.configurator.registry:AppRegistry")
    template_renderer_class: str = Field(
        default="fastlife.templating.renderer:JinjaxTemplateRenderer"
    )
    form_data_model_prefix: str = Field(default="payload")
    csrf_token_name: str = Field(default="csrf_token")

    session_secret_key: str = Field(default="")
    session_cookie_name: str = Field(default="flsess")
    session_cookie_domain: str = Field(default="")
    session_cookie_path: str = Field(default="/")
    session_duration: timedelta = Field(default=timedelta(days=14))
    session_cookie_same_site: Literal["lax", "strict", "none"] = Field(default="lax")
    session_cookie_secure: bool = Field(default=False)
    session_serializer: str = Field(
        default="fastlife.session.serializer:SignedSessionSerializer"
    )

    domain_name: str = Field(default="", title="domain name where the app is served")

    check_permission: str = Field(default="fastlife.security.policy:check_permission")



from typing import Optional

from .base import Widget


class BooleanWidget(Widget):
    def __init__(
        self,
        name: str,
        *,
        title: Optional[str],
        value: bool = False,
        removable: bool = False,
        token: str,
    ) -> None:
        super().__init__(name, title=title, removable=removable, token=token)
        self.value = value

    def get_template(self) -> str:
        return "pydantic_form.Boolean"



from typing import TYPE_CHECKING, Any, Mapping, Optional, Sequence, Type

from fastapi import Request
from jinjax.catalog import Catalog
from markupsafe import Markup
from pydantic.fields import FieldInfo

from fastlife.templating.renderer.widgets.factory import WidgetFactory

if TYPE_CHECKING:
    from fastlife.configurator.settings import Settings  # coverage: ignore

from fastlife.shared_utils.resolver import resolve_path

from .abstract import AbstractTemplateRenderer


def build_searchpath(template_search_path: str) -> Sequence[str]:
    searchpath: list[str] = []
    paths = template_search_path.split(",")

    for path in paths:
        if ":" in path:
            searchpath.append(resolve_path(path))
        else:
            searchpath.append(path)
    return searchpath


class JinjaxTemplateRenderer(AbstractTemplateRenderer):
    route_prefix: str
    """Used to prefix url to fetch fast life widgets."""

    def __init__(self, settings: "Settings") -> None:
        self.route_prefix = settings.fastlife_route_prefix
        self.form_data_model_prefix = settings.form_data_model_prefix
        self.csrf_token_name = settings.csrf_token_name

        self.catalog = Catalog()
        for path in build_searchpath(settings.template_search_path):
            self.catalog.add_folder(path)

    def render_page(self, request: Request, template: str, **params: Any) -> str:
        return self.catalog.render(  # type:ignore
            template,
            request=request,
            csrf_token={
                "name": self.csrf_token_name,
                "value": request.scope.get(self.csrf_token_name, ""),
            },
            pydantic_form=self.pydantic_form,
            **params
        )

    def render_template(self, template: str, **params: Any) -> str:
        return self.catalog.render(template, **params)  # type:ignore

    def pydantic_form(
        self,
        model: Type[Any],
        form_data: Optional[Mapping[str, Any]] = None,
        name: Optional[str] = None,
        token: Optional[str] = None,
        removable: bool = False,
        field: FieldInfo | None = None,
    ) -> Markup:
        return WidgetFactory(self, token).get_markup(
            model,
            form_data or {},
            prefix=(name or self.form_data_model_prefix),
            removable=removable,
            field=field,
        )



from typing import Optional, Sequence, Type, Union

from markupsafe import Markup
from pydantic import BaseModel

from fastlife.templating.renderer.abstract import AbstractTemplateRenderer

from .base import TypeWrapper, Widget


class UnionWidget(Widget):
    def __init__(
        self,
        name: str,
        *,
        title: Optional[str],
        child: Optional[Widget],
        children_types: Sequence[Type[BaseModel]],
        token: str,
        removable: bool,
    ):
        super().__init__(name, title=title, token=token, removable=removable)
        self.child = child
        self.children_types = children_types
        self.parent_name = name

    def build_types(self, route_prefix: str) -> Sequence[TypeWrapper]:
        return [
            TypeWrapper(typ, route_prefix, self.name, self.token)
            for typ in self.children_types
        ]

    def get_template(self) -> str:
        return "pydantic_form.Union"

    def to_html(self, renderer: "AbstractTemplateRenderer") -> Markup:
        """Return the html version"""
        child = Markup(self.child.to_html(renderer)) if self.child else ""
        return Markup(
            renderer.render_template(
                self.get_template(),
                widget=self,
                types=self.build_types(renderer.route_prefix),
                parent_type=TypeWrapper(
                    Union[tuple(self.children_types)],  # type: ignore
                    renderer.route_prefix,
                    self.parent_name,
                    self.token,
                    title=self.title,
                ),
                child=child,
            )
        )



import abc
from typing import Any, Mapping, Type

from fastapi import Request
from markupsafe import Markup
from pydantic.fields import FieldInfo


class AbstractTemplateRenderer(abc.ABC):
    route_prefix: str
    """Used to prefix url to fetch fast life widgets."""

    @abc.abstractmethod
    def render_page(self, request: Request, template: str, **params: Any) -> str:
        ...

    @abc.abstractmethod
    def render_template(self, template: str, **params: Any) -> str:
        ...

    @abc.abstractmethod
    def pydantic_form(
        self,
        model: Type[Any],
        form_data: Mapping[str, Any] | None = None,
        name: str | None = None,
        token: str | None = None,
        removable: bool = False,
        field: FieldInfo | None = None,
    ) -> Markup:
        ...



from typing import Optional, Sequence, Tuple

from .base import Widget


class DropDownWidget(Widget):
    def __init__(
        self,
        name: str,
        *,
        title: Optional[str],
        options: Sequence[Tuple[str, str]] | Sequence[str],
        removable: bool = False,
        value: str = "",
        token: Optional[str] = None,
        help_text: Optional[str] = None,
    ) -> None:
        super().__init__(name, title=title, token=token, removable=removable)
        self.options: list[dict[str, str]] = []
        for opt in options:
            if isinstance(opt, tuple):
                self.options.append({"value": opt[0], "text": opt[1]})
            else:
                self.options.append({"value": opt, "text": opt})
        self.value = value
        self.help_text = help_text

    def get_template(self) -> str:
        return "pydantic_form.Dropdown"



from datetime import timedelta
from typing import Literal, Type

from starlette.datastructures import MutableHeaders
from starlette.requests import HTTPConnection
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from fastlife.configurator.base import AbstractMiddleware

from .serializer import AbsractSessionSerializer, SignedSessionSerializer


class SessionMiddleware(AbstractMiddleware):
    def __init__(
        self,
        app: ASGIApp,
        cookie_name: str,
        secret_key: str,
        duration: timedelta,
        cookie_path: str = "/",
        cookie_same_site: Literal["lax", "strict", "none"] = "lax",
        cookie_secure: bool = False,
        cookie_domain: str = "",
        serializer: Type[AbsractSessionSerializer] = SignedSessionSerializer,
    ) -> None:
        self.app = app
        self.max_age = int(duration.total_seconds())
        self.serializer = serializer(secret_key, self.max_age)
        self.cookie_name = cookie_name
        self.security_flags = (
            f"Path={cookie_path}; HttpOnly; SameSite={cookie_same_site}"
        )
        if cookie_secure:
            self.security_flags += "; Secure"
        if cookie_domain:
            self.security_flags += f"; Domain={cookie_domain}"

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ("http", "websocket"):  # pragma: no cover
            await self.app(scope, receive, send)
            return
        connection = HTTPConnection(scope)
        existing_session = self.cookie_name in connection.cookies
        if existing_session:
            data = connection.cookies[self.cookie_name].encode("utf-8")
            scope["session"], broken_session = self.serializer.deserialize(data)
            if broken_session:
                existing_session = False
        else:
            scope["session"] = {}

        async def send_wrapper(message: Message) -> None:
            if message["type"] == "http.response.start":
                headers = None
                if scope["session"]:
                    # We have session data to persist.
                    data = self.serializer.serialize(scope["session"]).decode("utf-8")
                    headers = MutableHeaders(scope=message)
                    header_value = (
                        f"{self.cookie_name}={data}; "
                        f"Max-Age={self.max_age}; {self.security_flags}"
                    )
                    headers.append("set-cookie", header_value)
                elif existing_session:
                    # The session has been cleared.
                    headers = MutableHeaders(scope=message)
                    expires = "expires=Thu, 01 Jan 1970 00:00:00 GMT; "
                    header_value = (
                        f"{self.cookie_name}=; " f"{expires}{self.security_flags}"
                    )
                    headers.append("set-cookie", header_value)
            await send(message)

        await self.app(scope, receive, send_wrapper)



from typing import Any, Callable

from fastapi import Depends, Request, Response

from fastlife.configurator.registry import Registry
from fastlife.security.csrf import create_csrf_token

Template = Callable[..., Response]
TemplateEngine = Callable[["Registry", Request], Template]


def get_page_template(
    template: str, *, content_type: str = "text/html"
) -> TemplateEngine:
    def render_template(
        reg: "Registry",
        request: Request,
        *,
        _create_csrf_token: Callable[..., str] = create_csrf_token,
    ) -> Template:
        def parametrizer(**kwargs: Any) -> Response:
            request.scope[reg.settings.csrf_token_name] = (
                request.cookies.get(reg.settings.csrf_token_name)
                or _create_csrf_token()
            )
            data = reg.renderer.render_page(request, template, **kwargs)
            resp = Response(data, headers={"Content-Type": content_type})
            resp.set_cookie(
                reg.settings.csrf_token_name,
                request.scope[reg.settings.csrf_token_name],
                secure=request.url.scheme == "https",
                samesite="strict",
                max_age=60 * 15,
            )
            return resp

        return parametrizer

    return render_template


def get_template(
    template: str, *, content_type: str = "text/html"
) -> Callable[["Registry"], Template]:
    def render_template(reg: "Registry") -> Template:
        def parametrizer(**kwargs: Any) -> Response:
            data = reg.renderer.render_template(template, **kwargs)
            resp = Response(data, headers={"Content-Type": content_type})
            return resp

        return parametrizer

    return render_template


def template(template_path: str, page: bool = True) -> Template:
    return Depends(
        get_page_template(template_path) if page else get_template(template_path)
    )



from typing import Optional

from .base import Widget


class TextWidget(Widget):
    def __init__(
        self,
        name: str,
        *,
        title: Optional[str],
        aria_label: Optional[str] = None,
        placeholder: Optional[str] = None,
        removable: bool = False,
        value: str = "",
        token: Optional[str] = None,
        help_text: Optional[str] = None,
        input_type: str = "text"
    ) -> None:
        super().__init__(
            name, title=title, aria_label=aria_label, token=token, removable=removable
        )
        self.placeholder = placeholder or ""
        self.value = value
        self.help_text = help_text
        self.input_type = input_type

    def get_template(self) -> str:
        return "pydantic_form.Text"



import secrets
from collections.abc import MutableSequence, Sequence
from decimal import Decimal
from types import NoneType
from typing import Any, Literal, Mapping, Optional, Type, get_origin
from uuid import UUID

from markupsafe import Markup
from pydantic import BaseModel, EmailStr, SecretStr, ValidationError
from pydantic.fields import FieldInfo

from fastlife.shared_utils.infer import is_complex_type, is_union
from fastlife.templating.renderer.abstract import AbstractTemplateRenderer
from fastlife.templating.renderer.widgets.boolean import BooleanWidget
from fastlife.templating.renderer.widgets.dropdown import DropDownWidget
from fastlife.templating.renderer.widgets.hidden import HiddenWidget
from fastlife.templating.renderer.widgets.sequence import SequenceWidget

from .base import Widget, get_title
from .model import ModelWidget
from .text import TextWidget
from .union import UnionWidget


class WidgetFactory:
    def __init__(self, renderer: AbstractTemplateRenderer, token: Optional[str] = None):
        self.renderer = renderer
        self.token = token or secrets.token_urlsafe(4).replace("_", "-")

    def get_markup(
        self,
        base: Type[Any],
        form_data: Mapping[str, Any],
        *,
        prefix: str,
        removable: bool,
        field: FieldInfo | None = None,
    ) -> Markup:
        return self.get_widget(
            base, form_data, prefix=prefix, removable=removable, field=field
        ).to_html(self.renderer)

    def get_widget(
        self,
        base: Type[Any],
        form_data: Mapping[str, Any],
        *,
        prefix: str,
        removable: bool,
        field: FieldInfo | None = None,
    ) -> Widget:
        return self.build(
            base,
            value=form_data.get(prefix, {}),
            name=prefix,
            removable=removable,
            field=field,
        )

    def build(
        self,
        typ: Type[Any],
        *,
        name: str = "",
        value: Any,
        removable: bool,
        field: FieldInfo | None = None,
    ) -> Widget:
        type_origin = get_origin(typ)
        if type_origin:
            if is_union(typ):
                return self.build_union(name, typ, field, value, removable)

            if (
                type_origin is Sequence
                or type_origin is MutableSequence
                or type_origin is list
            ):
                return self.build_sequence(name, typ, field, value, removable)

            if type_origin is Literal:
                return self.build_literal(name, typ, field, value, removable)

        if issubclass(typ, BaseModel):  # if it raises here, the type_origin is unknown
            return self.build_model(name, typ, field, value or {}, removable)

        if issubclass(typ, bool):
            return self.build_boolean(name, typ, field, value or False, removable)

        if issubclass(typ, EmailStr):
            return self.build_emailtype(name, typ, field, value or "", removable)

        if issubclass(typ, SecretStr):
            return self.build_secretstr(name, typ, field, value or "", removable)

        if issubclass(typ, (int, str, float, Decimal, UUID)):
            return self.build_simpletype(name, typ, field, value or "", removable)

        raise NotImplementedError(f"{typ} not implemented")

    def build_model(
        self,
        field_name: str,
        typ: Type[BaseModel],
        field: Optional[FieldInfo],
        value: Mapping[str, Any],
        removable: bool,
    ) -> Widget:
        ret: dict[str, Any] = {}
        for key, field in typ.model_fields.items():
            child_key = f"{field_name}.{key}" if field_name else key
            if field.exclude:
                continue
            if field.annotation is None:
                raise ValueError(f"Missing annotation for {field} in {child_key}")
            ret[key] = self.build(
                field.annotation,
                name=child_key,
                field=field,
                value=value.get(key),
                removable=False,
            )
        return ModelWidget(
            field_name,
            children_widget=list(ret.values()),
            removable=removable,
            title=get_title(typ),
            token=self.token,
        )

    def build_union(
        self,
        field_name: str,
        field_type: Type[Any],
        field: Optional[FieldInfo],
        value: Any,
        removable: bool,
    ) -> Widget:
        types: list[Type[Any]] = []
        # required = True
        for typ in field_type.__args__:  # type: ignore
            if typ is NoneType:
                # required = False
                continue
            types.append(typ)  # type: ignore

        if (
            not removable
            and len(types) == 1
            # if the optional type is a complex type,
            and not is_complex_type(types[0])
        ):
            return self.build(
                types[0], name=field_name, field=field, value=value, removable=False
            )
        child = None
        if value:
            for typ in types:
                try:
                    typ(**value)
                except ValidationError:
                    pass
                else:
                    child = self.build(
                        typ,
                        name=field_name,
                        field=field,
                        value=value,
                        removable=False,
                    )

        widget = UnionWidget(
            field_name,
            # we assume those types are BaseModel
            child=child,
            children_types=types,  # type: ignore
            title=field.title if field else "",
            token=self.token,
            removable=removable,
        )

        return widget

    def build_sequence(
        self,
        field_name: str,
        field_type: Type[Any],
        field: Optional[FieldInfo],
        value: Optional[Sequence[Any]],
        removable: bool,
    ) -> Widget:
        typ = field_type.__args__[0]  # type: ignore
        value = value or []
        items = [
            self.build(
                typ,  # type: ignore
                name=f"{field_name}.{idx}",
                value=v,
                field=field,
                removable=True,
            )
            for idx, v in enumerate(value)
        ]
        return SequenceWidget(
            field_name,
            help_text=field.description if field else "",
            title=field.title if field else "",
            items=items,
            item_type=typ,  # type: ignore
            token=self.token,
            removable=removable,
        )

    def build_boolean(
        self,
        field_name: str,
        field_type: Type[Any],
        field: FieldInfo | None,
        value: bool,
        removable: bool,
    ) -> Widget:
        return BooleanWidget(
            field_name,
            removable=removable,
            title=field.title if field else "",
            token=self.token,
            value=value,
        )

    def build_emailtype(
        self,
        field_name: str,
        field_type: Type[Any],
        field: FieldInfo | None,
        value: str | int | float,
        removable: bool,
    ) -> Widget:
        return TextWidget(
            field_name,
            help_text=field.description if field else "",
            input_type="email",
            placeholder=str(field.examples[0]) if field and field.examples else None,
            removable=removable,
            title=field.title if field else "",
            token=self.token,
            value=str(value),
        )

    def build_secretstr(
        self,
        field_name: str,
        field_type: Type[Any],
        field: FieldInfo | None,
        value: SecretStr | str,
        removable: bool,
    ) -> Widget:
        return TextWidget(
            field_name,
            help_text=field.description if field else "",
            input_type="password",
            placeholder=str(field.examples[0]) if field and field.examples else None,
            removable=removable,
            title=field.title if field else "",
            token=self.token,
            value=value.get_secret_value() if isinstance(value, SecretStr) else value,
        )

    def build_literal(
        self,
        field_name: str,
        field_type: Type[Any],  # a literal actually
        field: FieldInfo | None,
        value: str | int | float,
        removable: bool,
    ) -> Widget:
        choices: list[str] = field_type.__args__  # type: ignore
        if len(choices) == 1:
            return HiddenWidget(
                field_name,
                value=choices[0],
                token=self.token,
            )
        return DropDownWidget(
            field_name,
            options=choices,
            removable=removable,
            title=field.title if field else "",
            token=self.token,
            value=str(value),
        )

    def build_simpletype(
        self,
        field_name: str,
        field_type: Type[Any],
        field: FieldInfo | None,
        value: str | int | float,
        removable: bool,
    ) -> Widget:
        return TextWidget(
            field_name,
            help_text=field.description if field else None,
            placeholder=str(field.examples[0]) if field and field.examples else None,
            aria_label=field.description if field else None,
            removable=removable,
            title=field.title if field else "",
            token=self.token,
            value=str(value),
        )



from typing import Sequence

from markupsafe import Markup

from fastlife.templating.renderer.abstract import AbstractTemplateRenderer

from .base import Widget


class ModelWidget(Widget):
    def __init__(
        self,
        name: str,
        *,
        children_widget: Sequence[Widget],
        removable: bool,
        title: str,
        token: str,
    ):
        super().__init__(name, title=title, removable=removable, token=token)
        self.children_widget = children_widget

    def get_template(self) -> str:
        return "pydantic_form.Model"

    def to_html(self, renderer: AbstractTemplateRenderer) -> Markup:
        """Return the html version"""
        children_widget = [child.to_html(renderer) for child in self.children_widget]
        kwargs = {
            "widget": self,
            "children_widget": children_widget,
        }
        return Markup(renderer.render_template(self.get_template(), **kwargs))



from .base import Widget


class HiddenWidget(Widget):
    def __init__(
        self,
        name: str,
        *,
        value: str,
        token: str,
    ) -> None:
        super().__init__(name, token=token)
        self.value = value

    def get_template(self) -> str:
        return "pydantic_form.Hidden"



import abc

from starlette.types import Receive, Scope, Send


class AbstractMiddleware(abc.ABC):
    @abc.abstractmethod
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        ...



from typing import Optional

from fastapi import Query, Response
from pydantic.fields import FieldInfo

from fastlife import Configurator, configure
from fastlife.configurator.registry import Registry
from fastlife.shared_utils.resolver import resolve_extended


async def show_widget(
    typ: str,
    reg: Registry,
    title: Optional[str] = Query(None),
    name: Optional[str] = Query(None),
    token: Optional[str] = Query(None),
    removable: bool = Query(False),
) -> Response:
    model_cls = resolve_extended(typ)
    field = None
    if title:
        field = FieldInfo(title=title)
    data = reg.renderer.pydantic_form(model_cls, None, name, token, removable, field)
    return Response(data, headers={"Content-Type": "text/html"})


@configure
def includeme(config: Configurator) -> None:
    route_prefix = config.registry.settings.fastlife_route_prefix
    config.add_route(
        "fl-pydantic-form-widget",
        f"{route_prefix}/pydantic-form/widgets/{{typ}}",
        show_widget,
        methods=["GET"],
    )



from typing import Any, Optional, Sequence, Type

from markupsafe import Markup

from fastlife.templating.renderer.abstract import AbstractTemplateRenderer

from .base import TypeWrapper, Widget


class SequenceWidget(Widget):
    items: Sequence[Widget]

    def __init__(
        self,
        name: str,
        *,
        title: Optional[str],
        help_text: Optional[str],
        items: Optional[Sequence[Widget]],
        item_type: Type[Any],
        token: str,
        removable: bool,
    ):
        super().__init__(name, title=title, token=token, removable=removable)
        self.items = items or []
        self.item_type = item_type
        self.help_text = help_text

    def get_template(self) -> str:
        return "pydantic_form/Sequence"

    def build_item_type(self, route_prefix: str) -> TypeWrapper:
        return TypeWrapper(self.item_type, route_prefix, self.name, self.token)

    def to_html(self, renderer: "AbstractTemplateRenderer") -> Markup:
        """Return the html version"""
        children = [Markup(item.to_html(renderer)) for item in self.items]
        return Markup(
            renderer.render_template(
                self.get_template(),
                widget=self,
                type=self.build_item_type(renderer.route_prefix),
                children_widgets=children,
            )
        )



import abc
import secrets
from typing import Any, Mapping, Optional, Type

from markupsafe import Markup

from fastlife.shared_utils.infer import is_union
from fastlife.templating.renderer.abstract import AbstractTemplateRenderer


def get_title(typ: Type[Any]) -> str:
    return getattr(
        getattr(typ, "__meta__", None),
        "title",
        getattr(typ, "__name__", ""),
    )


class Widget(abc.ABC):
    name: str
    "variable name, nested variables have dots"
    title: str
    "Human title for the widget"
    aria_label: str
    "Non visible text alternative"
    token: str
    "unique token to ensure id are unique in the DOM"
    removable: bool
    "Indicate that the widget is removable from the dom"

    def __init__(
        self,
        name: str,
        *,
        title: Optional[str] = None,
        token: Optional[str] = None,
        aria_label: Optional[str] = None,
        removable: bool = False,
    ):
        self.name = name
        self.title = title or name.split(".")[-1]
        self.aria_label = aria_label or ""
        self.token = token or secrets.token_urlsafe(4).replace("_", "-")
        self.removable = removable
        self.id = f"{self.name}-{self.token}".replace("_", "-").replace(".", "-")

    @abc.abstractmethod
    def get_template(self) -> str:
        ...

    def to_html(self, renderer: AbstractTemplateRenderer) -> Markup:
        """Return the html version"""
        return Markup(renderer.render_template(self.get_template(), widget=self))


def _get_fullname(typ: Type[Any]) -> str:
    if is_union(typ):
        typs = [_get_fullname(t) for t in typ.__args__]  # type: ignore
        return "|".join(typs)  # type: ignore
    return f"{typ.__module__}:{typ.__name__}"


class TypeWrapper:
    def __init__(
        self,
        typ: Type[Any],
        route_prefix: str,
        name: str,
        token: str,
        title: str | None = None,
    ):
        self.typ = typ
        self.route_prefix = route_prefix
        self.name = name
        self.title = title or get_title(typ)
        self.token = token

    @property
    def fullname(self) -> str:
        return _get_fullname(self.typ)

    @property
    def params(self) -> Mapping[str, str]:
        return {"name": self.name, "token": self.token, "title": self.title}

    @property
    def url(self) -> str:
        ret = f"{self.route_prefix}/pydantic-form/widgets/{self.fullname}"
        return ret



import importlib.util
from pathlib import Path
from typing import Any, Type, Union


def resolve(value: str) -> Any:
    """return the attr from the syntax: package.module:attr."""
    module_name, attr_name = value.split(":", 2)
    spec = importlib.util.find_spec(module_name)
    if spec is None:
        raise ValueError(f"Module {module_name} not found")
    module = importlib.util.module_from_spec(spec)
    if spec.loader is None:
        # I can't figure out when the spec.loader is None,
        # just relying on typing here
        raise ValueError("No loader on spec")  # coverage: ignore
    spec.loader.exec_module(module)

    try:
        attr = getattr(module, attr_name)
    except AttributeError:
        raise ValueError(f"Attribute {attr_name} not found in module {module_name}")

    return attr


def resolve_extended(value: str) -> Type[Any]:
    values = value.split("|")
    if len(values) == 1:
        return resolve(value)
    types = [resolve(t) for t in values if t != "builtins:NoneType"]
    return Union[tuple(types)]  # type: ignore


def resolve_path(value: str) -> str:
    package_name, resource_name = value.split(":", 1)
    spec = importlib.util.find_spec(package_name)
    if not spec or not spec.origin:
        raise ValueError(f"{value} not found")
    package_path = spec.origin
    full_path = Path(package_path).parent / resource_name
    return str(full_path)



from types import UnionType
from typing import Any, Type, Union, get_origin

from pydantic import BaseModel


def is_complex_type(typ: Type[Any]) -> bool:
    return get_origin(typ) or issubclass(typ, BaseModel)


def is_union(typ: Type[Any]) -> bool:
    type_origin = get_origin(typ)
    if type_origin:
        if type_origin is Union:  # Optional[T]
            return True

        if type_origin is UnionType:  # T | U
            return True
    return False



import re
import time
from collections.abc import MutableMapping
from typing import Any, Iterator, Literal, Mapping, Optional, Sequence
from urllib.parse import urlencode

import bs4
import httpx
from fastapi.testclient import TestClient
from starlette.types import ASGIApp

from fastlife.configurator.settings import Settings
from fastlife.session.serializer import AbsractSessionSerializer
from fastlife.shared_utils.resolver import resolve


class Element:
    def __init__(self, client: "WebTestClient", tag: bs4.Tag):
        self._client = client
        self._tag = tag

    def click(self) -> "WebResponse":
        return self._client.get(self._tag.attrs["href"])

    @property
    def node_name(self) -> str:
        return self._tag.name

    @property
    def attrs(self) -> dict[str, str]:
        return self._tag.attrs

    @property
    def text(self) -> str:
        return self._tag.text.strip()

    @property
    def form(self) -> "Element | None":
        return Element(self._client, self._tag.form) if self._tag.form else None

    @property
    def h1(self) -> "Element":
        nodes = self.by_node_name("h1")
        assert len(nodes) == 1, f"Should have 1 <h1>, got {len(nodes)} in {self}"
        return nodes[0]

    @property
    def h2(self) -> Sequence["Element"]:
        return self.by_node_name("h2")

    @property
    def hx_target(self) -> Optional[str]:
        el: bs4.Tag | None = self._tag
        while el:
            if "hx-target" in el.attrs:
                ret = el.attrs["hx-target"]
                if ret == "this":
                    ret = el.attrs["id"]
                return ret
            el = el.parent
        return None

    def by_text(self, text: str, *, node_name: str | None = None) -> "Element | None":
        nodes = self.iter_all_by_text(text, node_name=node_name)
        return next(nodes, None)

    def iter_all_by_text(
        self, text: str, *, node_name: str | None = None
    ) -> "Iterator[Element]":
        nodes = self._tag.find_all(string=re.compile(rf"\s*{text}\s*"))
        for node in nodes:
            if isinstance(node, bs4.NavigableString):
                node = node.parent

            if node_name:
                while node is not None:
                    if node.name == node_name:
                        yield Element(self._client, node)
                    node = node.parent
            elif node:
                yield Element(self._client, node)
        return None

    def get_all_by_text(
        self, text: str, *, node_name: str | None = None
    ) -> "Sequence[Element]":
        nodes = self.iter_all_by_text(text, node_name=node_name)
        return list(nodes)

    def by_label_text(self, text: str) -> "Element | None":
        label = self.by_text(text, node_name="label")
        assert label is not None
        assert label.attrs.get("for") is not None
        resp = self._tag.find(id=label.attrs["for"])
        assert not isinstance(resp, bs4.NavigableString)
        return Element(self._client, resp) if resp else None

    def by_node_name(
        self, node_name: str, *, attrs: dict[str, str] | None = None
    ) -> list["Element"]:
        return [
            Element(self._client, e) for e in self._tag.find_all(node_name, attrs or {})
        ]

    def __repr__(self) -> str:
        return f"<{self.node_name}>"

    def __str__(self) -> str:
        return str(self._tag)


class WebForm:
    def __init__(self, client: "WebTestClient", origin: str, form: Element):
        self._client = client
        self._form = form
        self._origin = origin
        self._formfields: dict[str, Element] = {}
        self._formdata: dict[str, str] = {}
        inputs = self._form.by_node_name("input")
        for input in inputs:
            self._formfields[input.attrs["name"]] = input
            if input.attrs.get("type") == "checkbox" and "checked" not in input.attrs:
                continue
            self._formdata[input.attrs["name"]] = input.attrs.get("value", "")

        inputs = self._form.by_node_name("select")
        for input in inputs:
            self._formfields[input.attrs["name"]] = input
            for option in input.by_node_name("options"):
                if "selected" in option.attrs:
                    self._formdata[input.attrs["name"]] = option.attrs.get(
                        "value", option.text
                    )
        # field textearea...

    def set(self, fieldname: str, value: str) -> Any:
        if fieldname not in self._formfields:
            raise ValueError(f"{fieldname} does not exists")
        if self._formfields[fieldname].node_name == "select":
            raise RuntimeError(f"{fieldname} is a <select>, use select() instead")
        self._formdata[fieldname] = value

    def select(self, fieldname: str, value: str) -> Any:
        if fieldname not in self._formfields:
            raise ValueError(f"{fieldname} does not exists")
        if self._formfields[fieldname].node_name != "select":
            raise RuntimeError(
                f"{fieldname} is a <{self._formfields[fieldname]}>, use set() instead"
            )
        if "multiple" in self._formfields[fieldname].attrs:
            raise NotImplementedError
        for option in self._formfields[fieldname].by_node_name("option"):
            if option.text == value.strip():
                self._formdata[fieldname] = option.attrs.get("value", option.text)
                break
        else:
            raise ValueError(f'No option {value} in <select name="{fieldname}">')

    def button(self, text: str, position: int = 0) -> "WebForm":
        buttons = self._form.get_all_by_text(text, node_name="button")
        assert len(buttons) > position, f'Button "{text}" not found'
        button = buttons[position]
        if "name" in button.attrs:
            self._formdata[button.attrs["name"]] = button.attrs.get("value", "")
        return self

    def submit(self, follow_redirects: bool = True) -> "WebResponse":
        headers = {}
        target = (
            self._form.attrs.get("hx-post")
            or self._form.attrs.get("post")
            or self._origin
        )
        if "hx-post" in self._form.attrs:
            if hx_target := self._form.hx_target:
                headers["HX-Target"] = hx_target

        return self._client.post(
            target,
            data=self._formdata,
            headers=headers,
            follow_redirects=follow_redirects,
        )

    def __contains__(self, key: str) -> bool:
        return key in self._formdata


class WebResponse:
    def __init__(self, client: "WebTestClient", origin: str, response: httpx.Response):
        self._client = client
        self._response = response
        self._origin = origin
        self._html: bs4.BeautifulSoup | None = None
        self._form: WebForm | None = None

    @property
    def status_code(self) -> int:
        return self._response.status_code

    @property
    def is_redirect(self) -> int:
        return 300 <= self._response.status_code < 400

    @property
    def content_type(self) -> str:
        return self._response.headers["content-type"]

    @property
    def headers(self) -> httpx.Headers:
        return self._response.headers

    @property
    def text(self) -> str:
        return self._response.text

    @property
    def html(self) -> Element:
        if self._html is None:
            self._html = bs4.BeautifulSoup(self._response.text, "html.parser")
        return Element(self._client, self._html)

    @property
    def html_body(self) -> Element:
        body = self.html.by_node_name("body")
        assert len(body) == 1
        return body[0]

    @property
    def form(self) -> WebForm:
        if self._form is None:
            form = self.html.form
            assert form is not None
            self._form = WebForm(self._client, self._origin, form)
        return self._form

    def by_text(self, text: str, *, node_name: str | None = None) -> Element | None:
        return self.html.by_text(text, node_name=node_name)

    def by_label_text(self, text: str) -> Element | None:
        return self.html.by_label_text(text)

    def by_node_name(
        self, node_name: str, *, attrs: dict[str, str] | None = None
    ) -> list[Element]:
        return self.html.by_node_name(node_name, attrs=attrs)


CookieTypes = httpx._types.CookieTypes  # type: ignore
Cookies = httpx._models.Cookies  # type: ignore


class Session(dict[str, Any]):
    def __init__(self, client: "WebTestClient"):
        self.client = client
        if client.session_serializer is None:
            raise RuntimeError(
                "WebTestClient has not been initialize with the app settings, "
                "can't decode session"
            )

        self.srlz = client.session_serializer
        settings = self.client.settings
        assert settings is not None
        self.settings = settings
        data: Mapping[str, Any]
        self.has_session = settings.session_cookie_name in self.client.cookies
        if self.has_session:
            data, error = self.srlz.deserialize(
                self.client.cookies[settings.session_cookie_name].encode("utf-8")
            )
            if error:
                self.has_session = False
        else:
            data = {}
        super().__init__(data)

    def __setitem__(self, __key: Any, __value: Any) -> None:
        super().__setitem__(__key, __value)
        settings = self.settings
        data = self.serialize()
        from http.cookiejar import Cookie

        self.client.cookies.jar.set_cookie(
            Cookie(
                version=0,
                name=settings.session_cookie_name,
                value=data,
                port=None,
                port_specified=False,
                domain=f".{settings.session_cookie_domain}",
                domain_specified=True,
                domain_initial_dot=True,
                path="/",
                path_specified=True,
                secure=False,
                expires=int(time.time() + settings.session_duration.total_seconds()),
                discard=False,
                comment=None,
                comment_url=None,
                rest={"HttpOnly": None, "SameSite": "lax"},  # type: ignore
                rfc2109=False,
            )
        )
        # this does not work
        # self.client.cookies.set(
        #     settings.session_cookie_name,
        #     data,
        #     settings.session_cookie_domain,
        #     settings.session_cookie_path,
        # )

    def serialize(self) -> str:
        return self.srlz.serialize(self).decode("utf-8")


class WebTestClient:
    def __init__(
        self,
        app: ASGIApp,
        *,
        settings: Settings | None = None,
        cookies: CookieTypes | None = None,
    ) -> None:
        self.app = app
        if settings is None:
            settings = Settings()
            settings.domain_name = settings.domain_name or "testserver.local"
        self.testclient = TestClient(
            app, base_url=f"http://{settings.domain_name}", cookies=cookies or {}
        )
        self.settings = settings
        self.session_serializer: AbsractSessionSerializer | None = None
        if settings:
            self.session_serializer = resolve(settings.session_serializer)(
                settings.session_secret_key,
                int(settings.session_duration.total_seconds()),
            )

    @property
    def cookies(self) -> Cookies:
        return self.testclient.cookies

    @cookies.setter
    def cookies(self, value: Cookies) -> None:
        self.testclient.cookies = value

    @property
    def session(self) -> MutableMapping[str, Any]:
        return Session(self)

    def request(
        self,
        method: Literal["GET", "POST"],  # I am a browser
        url: str,
        *,
        content: str | None = None,
        headers: Mapping[str, str] | None = None,
        max_redirects: int = 0,
    ) -> WebResponse:
        rawresp = self.testclient.request(
            method=method,
            url=url,
            headers=headers,
            content=content,
            follow_redirects=False,
        )
        resp = WebResponse(
            self,
            url,
            rawresp,
        )
        if resp.is_redirect and max_redirects > 0:
            if resp.status_code != 307:
                method = "GET"
                headers = None
                content = None
            return self.request(
                method=method,
                url=resp.headers["location"],
                content=content,
                headers=headers,
                max_redirects=max_redirects - 1,
            )
        if "HX-Redirect" in resp.headers and max_redirects > 0:
            return self.request(
                method="GET",
                url=resp.headers["HX-Redirect"],
                content=None,
                headers=headers,
                max_redirects=max_redirects - 1,
            )

        return resp

    def get(self, url: str, follow_redirects: bool = True) -> WebResponse:
        return self.request(
            "GET",
            url,
            max_redirects=int(follow_redirects) * 10,
        )

    def post(
        self,
        url: str,
        data: Mapping[str, Any],
        *,
        headers: Mapping[str, Any] | None = None,
        follow_redirects: bool = True,
    ) -> WebResponse:
        if headers is None:
            headers = {}
        return self.request(
            "POST",
            url,
            content=urlencode(data),
            headers={"Content-Type": "application/x-www-form-urlencoded", **headers},
            max_redirects=int(follow_redirects) * 10,
        )
